const as integer GRID_SIZE = 16
const as integer MAX_VOZES = 2

type canal
	private:
		buf as ALuint
		src as ALuint
	
	public:
		nota (0 to GRID_SIZE - 1) as _bool_
	
		declare sub inicia (nota as integer)
		declare sub reinicia
		declare sub toca
		declare sub destroi
end type

type sequenciador
	private:
		ch (0 to GRID_SIZE - 1) as canal
		cabecote as double
		cabecote_ant as double
		contador as integer
		
	public:
		declare sub inicia
		declare sub reinicia
		declare sub toca
		declare sub destroi
end type

type somBuffer
	public:
		snd as ALuint
	
		declare sub carrega (arquivo as string)
		declare sub destroi
end type

type somEngine
	private:
		voz (0 to MAX_VOZES - 1) as ALuint
		efeito as ALuint
		
		declare function outOfRange (i as integer) as _bool_
		
	public:	
		declare sub cria
		declare sub destroi
		
		declare sub startYourEngines
		declare sub para
		
		declare sub toca (buf as somBuffer, p as coord3d, vol as ALfloat)
		declare sub defineVoz (i as integer, buf as somBuffer)
		declare sub setPos (i as integer, p as coord3d)
		declare sub setVelocity (i as integer, v as coord3d)
		declare sub setVolume (i as integer, vol as ALfloat)
		declare sub setPitch (i as integer, pitch as ALfloat)
end type

'-----------------------------------------------------------------------
sub sequenciador.inicia
	ch(0).inicia 9
	ch(1).inicia 11
	ch(2).inicia 14
	ch(3).inicia 16
	ch(4).inicia 18
	ch(5).inicia 21
	ch(6).inicia 23
	ch(7).inicia 26
	ch(8).inicia 28
	ch(9).inicia 30
	ch(10).inicia 33
	ch(11).inicia 35
	ch(12).inicia 38
	ch(13).inicia 40
	ch(14).inicia 42
	ch(15).inicia 45
	
	cabecote = 0
	cabecote_ant = 0
	contador = 0
end sub

sub sequenciador.reinicia
	dim i as integer
	for i = 0 to GRID_SIZE - 1
		ch(i).reinicia
	next
end sub

sub sequenciador.destroi
	dim i as integer
	for i = 0 to GRID_SIZE - 1
		ch(i).destroi
	next
end sub

sub sequenciador.toca
	if not MUSICA_ON then exit sub

	dim as integer i, j

	'atualiza posição
	cabecote_ant = cabecote
	cabecote = cabecote + (0.18 * fps.getFator)

	'tem que tocar?
	if frac(cabecote_ant) > 0.5 and frac(cabecote) < 0.5 then
		'reinicia a sequência
		if cabecote >= GRID_SIZE then
			contador = contador + 1
			
			'limpa notas
			if contador >= 16 then
				dim as integer paridade = int(rnd * 2)
			
				for i = 0 to GRID_SIZE - 1
					for j = (i + paridade) mod 2 to GRID_SIZE - 1 step 2
						ch(i).nota(j) = false
					next j
				next i
				
				contador = 0
				
			'insere/remove uma nota
			else
				dim as integer x, y
				x = int(rnd * GRID_SIZE)
				y = int(rnd * GRID_SIZE)
				
				ch(x).nota(y) = not ch(x).nota(y)
			end if
			
			'volta cabeçote
			cabecote = cabecote - GRID_SIZE
		end if
		
		'toca
		for i = 0 to GRID_SIZE - 1
			if ch(i).nota(int(cabecote)) then ch(i).toca
		next i
	end if
end sub

'-----------------------------------------------------------------------
sub canal.inicia (nota as integer)
	'define frequência (nota = steps)
	dim as double freq = (261.626 / 2) * (2 ^ (nota / 12))
	
	'gera vozes
	alGenSources 1, @src
	alGenBuffers 1, @buf
	
	alSourcef src, AL_GAIN, (1 / GRID_SIZE) * 2
	alSource3f src, AL_POSITION, (((nota - 9) / 36) * 2) - 1, 0, 0
	
	'gera sample
	dim as uinteger taxa = 44100
	dim as integer tamanho = 1 * taxa 'segundos
	
	dim samples as short ptr
	samples = allocate(tamanho * sizeof(short)) '16 bits
	
	dim i as integer
	for i = 0 to tamanho - 1
		'gera onda
		dim as double passo = 2 * pi * (1 / taxa) * i
		dim as double ang = passo * freq
		dim as double fm = passo * 5900
		ang = ang + (sin(fm) * 0.1) 'carrier
		
		dim as double volume = ((tamanho - i) / tamanho)		
		volume = volume * ((2000 - freq) / 2000) 'equilibra
		
		samples[i] = 32760 * sin(ang) * volume
	next
	
	'alimenta buffer
	alSourcei src, AL_BUFFER, 0
	alBufferData buf, AL_FORMAT_MONO16, samples, tamanho * sizeof(short), taxa
	alSourcei src, AL_BUFFER, buf
	
	'libera ponteiro
	deallocate samples
	
	'inicia notas
	reinicia
end sub

sub canal.reinicia
	dim i as integer
	for i = 0 to GRID_SIZE - 1
		this.nota(i) = false
	next
end sub

sub canal.destroi
	alDeleteBuffers 1, @buf
	alDeleteSources 1, @src
end sub

sub canal.toca
	alSourceStop src
	alSourcePlay src
end sub

'-----------------------------------------------------------------------
sub somBuffer.carrega (arquivo as string)
	snd = alutCreateBufferFromFile(arquivo)
end sub

sub somBuffer.destroi
	alDeleteBuffers 1, @snd
end sub

'-----------------------------------------------------------------------

function somEngine.outOfRange (i as integer) as _bool_
	if i < 0 or i > MAX_VOZES - 1 then
		return true
	else
		return false
	end if
end function

sub somEngine.cria
	dim i as integer
	
	'cria sources
	alGenSources MAX_VOZES, @voz(0)
	alGenSources 1, @efeito
	
	'define propriedades
	for i = 0 to MAX_VOZES - 1
		alSourcei voz(i), AL_LOOPING, AL_TRUE
		alSourcef voz(i), AL_PITCH, 1.0
		alSourcef voz(i), AL_GAIN, 1.0
		alSourcei voz(i), AL_REFERENCE_DISTANCE, 20
		alSourcei voz(i), AL_ROLLOFF_FACTOR, 0.95
	next i

	alSourcei efeito, AL_LOOPING, AL_FALSE
	alSourcei efeito, AL_PITCH, 1.0
	alSourcei efeito, AL_GAIN, 1.0
	alSourcei efeito, AL_REFERENCE_DISTANCE, 20
	alSourcei efeito, AL_ROLLOFF_FACTOR, 0.95
end sub

sub somEngine.destroi
	dim i as integer
	
	'desliga buffers
	for i = 0 to MAX_VOZES
		alSourcei voz(i), AL_BUFFER, NULL
	next
	
	alSourcei efeito, AL_BUFFER, NULL
	
	'remove sources
	alDeleteSources MAX_VOZES, @voz(0)
	alDeleteSources 1, @efeito
end sub

sub somEngine.startYourEngines
	dim i as integer
	
	if not SFX_ON then exit sub
	
	for i = 0 to MAX_VOZES - 1
		alSourcePlay voz(i)
	next i
end sub

sub somEngine.para
	dim i as integer
	
	for i = 0 to MAX_VOZES - 1
		alSourceStop voz(i)
	next i
end sub

sub somEngine.toca (buf as somBuffer, p as coord3d, vol as ALfloat)
	'pode tocar?
	if not SFX_ON then exit sub

	'já está tocando?
	dim estado as ALenum
	
	alGetSourcei efeito, AL_SOURCE_STATE, @estado
	if not estado = AL_PLAYING then
		'define buffer
		alSourcei efeito, AL_BUFFER, buf.snd
		alSource3f efeito, AL_POSITION, p.x, p.y, p.z
		alSourcef efeito, AL_GAIN, vol
		
		'toca
		alSourcePlay efeito
	end if
end sub

sub somEngine.defineVoz (i as integer, buf as somBuffer)
	if outOfRange(i) then exit sub
	alSourcei voz(i), AL_BUFFER, buf.snd
end sub

sub somEngine.setPos (i as integer, p as coord3d)
	if outOfRange(i) then exit sub
	alSource3f voz(i), AL_POSITION, p.x, p.y, p.z
end sub

sub somEngine.setVelocity (i as integer, v as coord3d)
	if outOfRange(i) then exit sub
	alSource3f voz(i), AL_VELOCITY, v.x, v.y, v.z 
end sub

sub somEngine.setVolume (i as integer, vol as ALfloat)
	if outOfRange(i) then exit sub
	alSourcef voz(i), AL_GAIN, vol
end sub

sub somEngine.setPitch (i as integer, pitch as ALfloat)
	if outOfRange(i) then exit sub
	alSourcef voz(i), AL_PITCH, pitch
end sub
